/***********************************************************************

Vorosketch User Distances Demo by Herman Haverkort, 3 March 2023

************************************************************************

Copyright 2023 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file is designed to be used with Vorosketch v0.31 BETA; it might 
work with later versions too. New versions may be available from 
http://herman.haverkort.net.


***********************************************************************/

// This file defines some distances as examples of what could be user-
// defined distances, including some distances that might be included 
// into Vorosketch proper later but which are still subject to change.

// When Vorosketch is compiled with the -D INCLUDE_USER_DISTANCES option,
// the following macro is included to register distance definitions from
// this file with the parser for the -m option:

#define MAKEUSERDISTANCE \
if (dname == "mydistance") return new MyDistance();\
if (dname == "@mydistance") return new cMyDistance();\
if (dname == "wave") return new WaveDistance();\
if (dname == "@wiggle") return new cWiggleDistance();\
if (dname == "poincare halfplane") return new PoincareHalfplaneDistance();\
if (dname == "inverted"||\
    dname == "inverted euclidean") return new InvertedEuclideanDistance();\
if (dname == "disk"||\
    dname == "euclidean disk") return new EuclideanDiskDistance();


class MyDistance: public PointDistance
{
    public:
  virtual const string description() const
  {
    return "mydistance";
  }

  virtual DOUBLE distanceToPointSite(const Point& p, const PointSite* s) const
  {
    return (p - s->location).length();
  }
};


class cMyDistance: public Distance
{
    public:
  virtual const string description() const
  {
    return "@mydistance";
  }

  virtual DOUBLE distanceToSite(const Point& p, const Site* s) const
  {
    return p.length();
  }
};


class WaveDistance: public SetDistance
{
    public:
  virtual const string description() const
  {
    return "wave";
  }

  virtual DOUBLE distanceToPointSite(const Point& p, const PointSite* s) const
  {
    DOUBLE d = EuclideanDistance().distanceToPointSite(p, s);
    return(1.0 + (sin(d * 10 - PI * 0.5) - 1.0) * pow(0.1, d) * 0.5);
  }

  virtual DOUBLE distanceToSegmentSite(const Point& p, const SegmentSite* s) const
  {
    DOUBLE d = EuclideanDistance().distanceToSegmentSite(p, s);
    return(1.0 + (sin(d * 10 - PI * 0.5) - 1.0) * pow(0.1, d) * 0.5);
  }
};


class cWiggleDistance: public FieldDistance
{
    public:
  virtual const string description() const
  {
    return "@wiggle";
  }

  virtual DOUBLE distanceToSite(const Point& p, const Site* s) const
  {
    DOUBLE b = (p.inOrigin() ? 0 : atan2(p.y, p.x));
    DOUBLE wiggle = 0;
    for (int i = 4; i <= 32; i *= 4) wiggle += sin(b * i + HALFPI/i)/i;
    for (int i = 8; i <= 32; i *= 4) wiggle -= sin(b * i + HALFPI/i)/i;
    return wiggle;
  }

  virtual void boundDistanceToSite(
    const Point& centre, DOUBLE halfwidth, const Site* s, Range& range
  ) const
  {
    range.set(-0.5,0.5);
  }
};


class PoincareHalfplaneDistance: public PointDistance
{
    public:
  virtual const string description() const
  {
    return "poincare halfplane";
  }

  virtual DOUBLE distanceToPointSite(const Point& p, const PointSite* s) const
  {
    if (p.y <= -1 || s->location.y <= -1) return NAN;
    // take y-coordinates +1 to put the bottom line at y = -1:
    return asinh((p - s->location).length() * 0.5 / sqrt((p.y+1)*(s->location.y+1))) * 2.0;
  }
};


class InvertedEuclideanDistance: public PointDistance
{
    public:
  virtual const string description() const
  {
    return "inverted euclidean";
  }

  virtual DOUBLE distanceToPointSite(const Point& p, const PointSite* s) const
  {
    if (p.inOrigin() || s->location.inOrigin()) return INF;
    return (p / p.sqr() - s->location / s->location.sqr()).length();
  }
};


class EuclideanDiskDistance: public PointDistance
{
    public:
  virtual const string description() const
  {
    return "euclidean disk";
  }

  virtual DOUBLE distanceToPointSite(const Point& p, const PointSite* s) const
  {
    if (p.length() >= 1 || s->distanceFromOrigin >= 1) return NAN;
    return (p / ((DOUBLE) 1.0 - p.length()) - s->location / ((DOUBLE) 1.0 - s->distanceFromOrigin)).length();
  }
};




